from django import forms
from django.contrib import admin
from django_admin_hstore_widget.forms import HStoreFormField
from enumfields.admin import EnumFieldListFilter

from arkindex.documents.models import (
    AllowedMetaData,
    Classification,
    Corpus,
    Element,
    ElementType,
    Entity,
    EntityLink,
    EntityRole,
    MetaData,
    MLClass,
    Transcription,
)
from arkindex.users.admin import GroupMembershipInline, UserMembershipInline


class ElementTypeInline(admin.TabularInline):
    model = ElementType
    prepopulated_fields = {'slug': ('display_name', )}


class CorpusAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'public', 'repository', 'top_level_type', 'created')
    raw_id_fields = ('thumbnail', )
    search_fields = ('name', )
    inlines = (ElementTypeInline, UserMembershipInline, GroupMembershipInline)
    ordering = ('-created', )

    def has_delete_permission(self, request, obj=None):
        # Require everyone to use the asynchronous corpus deletion
        return False

    def get_queryset(self, request):
        return super().get_queryset(request).select_related('top_level_type')

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        if obj:
            # Limit top level type queryset to the types of this corpus
            form.base_fields['top_level_type'].queryset = ElementType.objects.filter(corpus=obj)
        return form


class ClassificationInline(admin.TabularInline):
    model = Classification
    readonly_fields = ('confidence', 'high_confidence', )
    raw_id_fields = ('worker_version', 'moderator', 'ml_class')


class AllowedMetaDataAdmin(admin.ModelAdmin):
    list_display = ('id', 'corpus', 'type', 'name')
    readonly_fields = ('id', )

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        form.base_fields['corpus'].queryset = Corpus.objects.order_by('name', 'id')
        return form


class MetaDataAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'type', )
    list_filter = [('type', EnumFieldListFilter), ]
    readonly_fields = ('id', )
    raw_id_fields = ('element', 'entity', )


class MetaDataInline(admin.TabularInline):
    model = MetaData
    raw_id_fields = ('entity', 'worker_version', )


class ElementAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'type', 'corpus', )
    list_filter = ['type__slug', 'corpus']
    fields = ('id', 'type', 'name', 'corpus')
    readonly_fields = ('id', 'corpus', )
    search_fields = ('name', )
    inlines = (MetaDataInline, ClassificationInline)

    # Disable element creation through the admin to prevent element_type conflicts
    def has_add_permission(self, request):
        return False

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == 'type':
            # Only display types available on the element's corpus
            # It would make no sense to pick a type from another corpus
            # Display all elements on creation, as we do not know yet the corpus
            element_id = request.resolver_match.kwargs.get('object_id')
            if element_id:
                element = self.get_object(request, element_id)
                kwargs['queryset'] = ElementType.objects.filter(corpus=element.corpus).order_by('display_name')

        return super().formfield_for_foreignkey(db_field, request, **kwargs)


class TranscriptionAdmin(admin.ModelAdmin):
    list_display = ('id', 'text', 'confidence', 'element', )
    fields = ('id', 'text', 'confidence', 'element', )
    readonly_fields = ('id', )
    raw_id_fields = ('element', )


class MLClassAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'corpus')
    list_filter = ('corpus',)
    search_fields = ('name',)
    fields = ('name', 'corpus')

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        form.base_fields['corpus'].queryset = Corpus.objects.order_by('name', 'id')
        return form


class EntityMetaForm(forms.ModelForm):
    metas = HStoreFormField()


class EntityLinkInLine(admin.TabularInline):
    model = EntityLink
    fk_name = 'parent'
    raw_id_fields = ('child', )


class EntityAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'type')
    list_filter = [('type', EnumFieldListFilter), 'corpus']
    readonly_fields = ('id', )
    search_fields = ('name', )
    inlines = (EntityLinkInLine, )
    form = EntityMetaForm


class EntityRoleAdmin(admin.ModelAdmin):
    list_display = ('id', 'corpus', 'parent_name', 'child_name')
    list_filter = ('corpus', )
    readonly_fields = ('id', )
    ordering = ('corpus', 'parent_name', 'child_name')


admin.site.register(Corpus, CorpusAdmin)
admin.site.register(Element, ElementAdmin)
admin.site.register(Transcription, TranscriptionAdmin)
admin.site.register(MLClass, MLClassAdmin)
admin.site.register(MetaData, MetaDataAdmin)
admin.site.register(Entity, EntityAdmin)
admin.site.register(EntityRole, EntityRoleAdmin)
admin.site.register(AllowedMetaData, AllowedMetaDataAdmin)
