PRAGMA foreign_keys = ON;

CREATE TABLE image (
    id VARCHAR(37) NOT NULL,
    url TEXT NOT NULL,
    width INTEGER NOT NULL,
    height INTEGER NOT NULL,
    PRIMARY KEY (id),
    UNIQUE (url),
    CHECK (width >= 0 AND height >= 0)
);

CREATE TABLE worker_version (
    id VARCHAR(37) NOT NULL,
    name VARCHAR(100) NOT NULL,
    slug VARCHAR(100) NOT NULL,
    type VARCHAR(50) NOT NULL,
    revision VARCHAR(50) NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE element (
    id VARCHAR(37) NOT NULL,
    created REAL NOT NULL,
    updated REAL NOT NULL,
    name VARCHAR(250) NOT NULL,
    type VARCHAR(50) NOT NULL,
    polygon TEXT,
    image_id VARCHAR(37),
    worker_version_id VARCHAR(37),
    PRIMARY KEY (id),
    FOREIGN KEY (image_id) REFERENCES image (id) ON DELETE CASCADE,
    FOREIGN KEY (worker_version_id) REFERENCES worker_version (id) ON DELETE CASCADE
);

CREATE TABLE element_path (
    id VARCHAR(37) NOT NULL,
    parent_id VARCHAR(37) NOT NULL,
    child_id VARCHAR(37) NOT NULL,
    ordering INTEGER NOT NULL DEFAULT 0,
    PRIMARY KEY (id),
    FOREIGN KEY (parent_id) REFERENCES element (id) ON DELETE CASCADE,
    FOREIGN KEY (child_id) REFERENCES element (id) ON DELETE CASCADE,
    UNIQUE (parent_id, child_id),
    CHECK (ordering >= 0)
);

CREATE TABLE transcription (
    id VARCHAR(37) NOT NULL,
    element_id VARCHAR(37) NOT NULL,
    text TEXT NOT NULL,
    confidence REAL,
    worker_version_id VARCHAR(37),
    PRIMARY KEY (id),
    FOREIGN KEY (element_id) REFERENCES element (id) ON DELETE CASCADE,
    FOREIGN KEY (worker_version_id) REFERENCES worker_version (id) ON DELETE CASCADE,
    CHECK (confidence IS NULL OR (confidence >= 0 AND confidence <= 1))
);

CREATE TABLE classification (
    id VARCHAR(37) NOT NULL,
    element_id VARCHAR(37) NOT NULL,
    class_name VARCHAR(1024) NOT NULL,
    state VARCHAR(16) NOT NULL DEFAULT 'pending',
    moderator VARCHAR(255),
    confidence REAL,
    high_confidence INTEGER NOT NULL DEFAULT 0,
    worker_version_id VARCHAR(37),
    PRIMARY KEY (id),
    FOREIGN KEY (element_id) REFERENCES element (id) ON DELETE CASCADE,
    FOREIGN KEY (worker_version_id) REFERENCES worker_version (id) ON DELETE CASCADE,
    CHECK (confidence IS NULL OR (confidence >= 0 AND confidence <= 1)),
    CHECK (high_confidence = 0 OR high_confidence = 1)
);

CREATE TABLE entity (
    id VARCHAR(37) NOT NULL,
    name TEXT NOT NULL,
    type VARCHAR(50) NOT NULL,
    validated INTEGER NOT NULL DEFAULT 0,
    moderator VARCHAR(255),
    metas TEXT,
    worker_version_id VARCHAR(37),
    PRIMARY KEY (id),
    FOREIGN KEY (worker_version_id) REFERENCES worker_version (id) ON DELETE CASCADE,
    CHECK (validated = 0 OR validated = 1)
);

CREATE TABLE transcription_entity (
    id VARCHAR(37) NOT NULL,
    transcription_id VARCHAR(37) NOT NULL,
    entity_id VARCHAR(37) NOT NULL,
    offset INTEGER NOT NULL,
    length INTEGER NOT NULL,
    worker_version_id VARCHAR(37),
    PRIMARY KEY (id),
    FOREIGN KEY (transcription_id) REFERENCES transcription (id) ON DELETE CASCADE,
    FOREIGN KEY (entity_id) REFERENCES entity (id) ON DELETE CASCADE,
    FOREIGN KEY (worker_version_id) REFERENCES worker_version (id) ON DELETE CASCADE,
    UNIQUE (transcription_id, entity_id, offset, length, worker_version_id),
    CHECK (offset >= 0 AND length >= 0)
);

CREATE TABLE entity_role (
    id VARCHAR(37) NOT NULL,
    parent_name VARCHAR(250) NOT NULL,
    child_name VARCHAR(250) NOT NULL,
    parent_type VARCHAR(50) NOT NULL,
    child_type VARCHAR(50) NOT NULL,
    PRIMARY KEY (id),
    UNIQUE (parent_name, child_name, parent_type, child_type)
);

CREATE TABLE entity_link (
    id VARCHAR(37) NOT NULL,
    parent_id VARCHAR(37) NOT NULL,
    child_id VARCHAR(37) NOT NULL,
    role_id VARCHAR(37) NOT NULL,
    PRIMARY KEY (id),
    FOREIGN KEY (parent_id) REFERENCES entity (id),
    FOREIGN KEY (child_id) REFERENCES entity (id),
    FOREIGN KEY (role_id) REFERENCES entity_role (id)
);

CREATE TABLE metadata (
    id VARCHAR(37) NOT NULL,
    element_id VARCHAR(37) NOT NULL,
    name VARCHAR(250) NOT NULL,
    type VARCHAR(50) NOT NULL,
    value TEXT NOT NULL,
    entity_id VARCHAR(37),
    worker_version_id VARCHAR(37),
    PRIMARY KEY (id),
    FOREIGN KEY (element_id) REFERENCES element (id) ON DELETE CASCADE,
    FOREIGN KEY (entity_id) REFERENCES entity (id) ON DELETE SET NULL,
    FOREIGN KEY (worker_version_id) REFERENCES worker_version (id) ON DELETE CASCADE
);
