import math

from django.conf import settings
from rest_framework import serializers

from arkindex.documents.date_parser import parse_date
from arkindex.documents.models import Element, ElementType, Entity, EntityType, MetaType
from arkindex.documents.serializers.light import CorpusLightSerializer, ElementLightSerializer
from arkindex.documents.serializers.ml import TranscriptionSerializer
from arkindex.images.serializers import ZoneSerializer
from arkindex.project.serializer_fields import EnumField


class SearchQuerySerializer(serializers.Serializer):
    """
    Search parameters validation serializer in order to build a ES query

    date_lte is rounded to superior value using ES syntax (see round_date()) and served as date_lt (exactly lower than)
    Rounded value depend if month/day are provided (Ex: '<= march 1999' becomes '< april 1999')
    """
    q = serializers.CharField(source='query', default=None, max_length=settings.ES_QUERY_STRING_MAX_LENGTH)
    confidence = serializers.FloatField(source='min_confidence', min_value=0.0, max_value=1.0, default=0.0)
    date_gte = serializers.CharField(default=None)
    date_lte = serializers.CharField(source='date_lt', default=None)
    element_type = serializers.SlugField(default=None)
    corpus = serializers.UUIDField(source='corpus_id', default=None)

    def parse_date(self, raw_date):
        date = parse_date(raw_date)
        if len(date) != 1:
            raise serializers.ValidationError('Could not parse Date')
        return date[0]

    def validate_confidence(self, value):
        if not isinstance(value, float) or math.isinf(value) or math.isnan(value):
            raise serializers.ValidationError('Confidence should be a valid decimal number between 0 and 1')
        return value

    def validate_date_gte(self, value):
        if not value:
            return
        return self.parse_date(value).es_str()

    def validate_date_lte(self, value):
        if not value:
            return
        return self.parse_date(value).es_round_up()

    def validate(self, data):
        data = super().validate(data)

        gte, lt = data['date_gte'], data['date_lt']
        if gte and lt and gte > lt:
            raise serializers.ValidationError("Upper date must be greater than lower date")
        if not (data['query'] or data['date_lt'] or data['date_gte']):
            raise serializers.ValidationError("At least query terms or dates must be provided")

        corpus_id, element_type = data['corpus_id'], data['element_type']
        if element_type:
            if not corpus_id:
                raise serializers.ValidationError('A corpus is required to filter by element types')
            if not ElementType.objects.filter(corpus_id=corpus_id, slug=element_type).exists():
                raise serializers.ValidationError(
                    'Type with slug {!r} does not exist in corpus {}'.format(element_type, corpus_id),
                )

        return data


class EntitySearchQuerySerializer(serializers.Serializer):
    """
    Search parameters validation serializer in order to build a ES query for entity
    """
    q = serializers.CharField(source='query', max_length=settings.ES_QUERY_STRING_MAX_LENGTH)
    type = EnumField(enum=EntityType, default=None)
    corpus = serializers.UUIDField(source='corpus_id', default=None)


class IIIFSearchQuerySerializer(serializers.Serializer):
    """
    Search parameters for IIIF transcription search
    See https://iiif.io/api/search/1.0/#request-1
    """
    q = serializers.CharField(source='query', max_length=settings.ES_QUERY_STRING_MAX_LENGTH)


class ElementSearchResultSerializer(ElementLightSerializer):
    """
    A page search result with nested transcriptions
    """
    zone = ZoneSerializer()
    transcriptions = TranscriptionSerializer(many=True, source='transcriptions_results')
    total_transcriptions = serializers.IntegerField()
    parent_paths = serializers.ListField(
        child=serializers.ListField(
            child=ElementLightSerializer()
        ),
    )
    corpus = CorpusLightSerializer()
    date_from = serializers.CharField()
    date_to = serializers.CharField()

    class Meta:
        model = Element
        fields = (
            'id',
            'name',
            'type',
            'zone',
            'transcriptions',
            'total_transcriptions',
            'parent_paths',
            'corpus',
            'date_from',
            'date_to',
        )


class EntitySearchResultSerializer(serializers.ModelSerializer):
    """
    Serializes an entity
    """
    type = EnumField(EntityType)

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas'
        )


class SolrDocumentSerializer(serializers.Serializer):
    """
    Serializes a Solr document
    """
    id = serializers.UUIDField()

    parent_id = serializers.UUIDField()
    parent_name = serializers.CharField()
    parent_type = serializers.CharField()

    element_id = serializers.UUIDField()
    element_text = serializers.CharField()
    element_type = serializers.CharField()
    element_worker = serializers.CharField(allow_null=True)
    element_image = serializers.URLField(allow_null=True)

    transcription_id = serializers.UUIDField(allow_null=True)
    transcription_text = serializers.CharField(allow_null=True)
    transcription_confidence = serializers.FloatField(min_value=0, max_value=1, required=False)
    transcription_worker = serializers.CharField(allow_null=True)

    classification_id = serializers.UUIDField(allow_null=True)
    classification_name = serializers.CharField(allow_null=True)
    classification_confidence = serializers.FloatField(min_value=0, max_value=1, required=False)
    classification_worker = serializers.CharField(allow_null=True)

    metadata_id = serializers.UUIDField(allow_null=True)
    metadata_name = serializers.CharField(allow_null=True)
    metadata_text = serializers.CharField(allow_null=True)
    metadata_type = EnumField(MetaType, allow_null=True)
    metadata_worker = serializers.CharField(allow_null=True)

    entity_id = serializers.UUIDField(allow_null=True)
    entity_text = serializers.CharField(allow_null=True)
    entity_type = EnumField(EntityType, allow_null=True)
    entity_worker = serializers.CharField(allow_null=True)


class SolrFacetsSerializer(serializers.Serializer):
    """
    Serializes Solr facets
    """
    element_type = serializers.DictField()
    element_worker = serializers.DictField()

    transcription_worker = serializers.DictField()

    classification_name = serializers.DictField()
    classification_worker = serializers.DictField()

    metadata_name = serializers.DictField()
    metadata_type = serializers.DictField()
    metadata_worker = serializers.DictField()

    entity_type = serializers.DictField()
    entity_worker = serializers.DictField()


class CorpusSearchResultSerializer(serializers.Serializer):
    """
    Serializes Corpus Search results
    """
    count = serializers.IntegerField()
    number = serializers.IntegerField()
    next = serializers.URLField(allow_null=True)
    previous = serializers.URLField(allow_null=True)
    results = SolrDocumentSerializer(many=True, allow_null=True)
    facets = SolrFacetsSerializer()


class CorpusSearchQuerySerializer(serializers.Serializer):
    """
    Search parameters validation serializer in order to build a Solr query
    """
    query = serializers.CharField()
    sources = serializers.MultipleChoiceField([
        ('element', 'element'),
        ('transcription', 'transcription'),
        ('metadata', 'metadata'),
        ('entity', 'entity')
    ], default={'element', 'transcription', 'metadata', 'entity'})
    only_facets = serializers.BooleanField(default=False)
    page = serializers.IntegerField(default=1, min_value=1)

    # Facets
    element_type = serializers.CharField(required=False)
    element_worker = serializers.CharField(required=False)

    transcription_worker = serializers.CharField(required=False)

    classification_name = serializers.CharField(required=False)
    classification_worker = serializers.CharField(required=False)

    metadata_name = serializers.CharField(required=False)
    metadata_type = serializers.CharField(required=False)
    metadata_worker = serializers.CharField(required=False)

    entity_type = serializers.CharField(required=False)
    entity_worker = serializers.CharField(required=False)
