import json
from unittest.mock import patch

from django.db.models import CharField, Value

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.indexer_v2 import Indexer
from arkindex.documents.models import Corpus, EntityType, MetaType
from arkindex.images.models import Zone
from arkindex.project.tests import FixtureTestCase


class TestIndexerV2Command(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name='private', indexable=True)
        cls.worker_version = WorkerVersion.objects.first()
        cls.worker = cls.worker_version.worker
        page_type, _ = cls.private_corpus.types.get_or_create(slug='page', display_name='Page', indexable=True)
        cls.page = cls.private_corpus.elements.create(name='New page', type=page_type, worker_version=cls.worker_version)

    @patch('arkindex.documents.indexer_v2.solr')
    def test_setup(self, mock_solr):
        mock_collections = mock_solr.collections
        mock_schema = mock_solr.schema
        mock_collections.exists.return_value = False
        mock_schema.does_field_exist.return_value = False

        indexer = Indexer('corpus_id')
        indexer.setup()

        self.assertEqual(mock_collections.exists.call_count, 1)
        (index_name, ), _ = mock_collections.exists.call_args
        self.assertEqual(index_name, indexer.collection_name)

        self.assertEqual(mock_collections.create.call_count, 1)
        (index_name, args), _ = mock_collections.create.call_args
        self.assertEqual(index_name, indexer.collection_name)
        self.assertEqual(args, indexer.solr_num_shards)

        self.assertEqual(mock_solr.transport.send_request.call_count, len(indexer.solr_type_fields))
        for type_field, (_, kwargs) in zip(indexer.solr_type_fields, mock_solr.transport.send_request.call_args_list):
            self.assertDictEqual(kwargs, {
                'method': 'POST',
                'endpoint': mock_solr.schema.schema_endpoint,
                'collection': indexer.collection_name,
                'data': json.dumps({'add-field-type': type_field})
            })

        self.assertEqual(mock_schema.does_field_exist.call_count, len(indexer.solr_fields))
        self.assertEqual(mock_schema.create_field.call_count, len(indexer.solr_fields))
        for field, ((index_name, args), _) in zip(indexer.solr_fields, mock_schema.create_field.call_args_list):
            self.assertEqual(index_name, indexer.collection_name)
            self.assertDictEqual(args, field)
        self.assertEqual(mock_schema.replace_field.call_count, 0)

    @patch('arkindex.documents.indexer_v2.solr')
    def test_already_setup(self, mock_solr):
        mock_collections = mock_solr.collections
        mock_schema = mock_solr.schema
        mock_collections.exists.return_value = True
        mock_schema.does_field_exist.return_value = True

        indexer = Indexer('corpus_id')
        indexer.setup()

        self.assertEqual(mock_collections.exists.call_count, 1)
        (index_name, ), _ = mock_collections.exists.call_args
        self.assertEqual(index_name, indexer.collection_name)

        self.assertEqual(mock_collections.create.call_count, 0)

        self.assertEqual(mock_solr.transport.send_request.call_count, len(indexer.solr_type_fields))
        for type_field, (_, kwargs) in zip(indexer.solr_type_fields, mock_solr.transport.send_request.call_args_list):
            self.assertDictEqual(kwargs, {
                'method': 'POST',
                'endpoint': mock_solr.schema.schema_endpoint,
                'collection': indexer.collection_name,
                'data': json.dumps({'add-field-type': type_field})
            })

        self.assertEqual(mock_schema.does_field_exist.call_count, len(indexer.solr_fields))
        self.assertEqual(mock_schema.replace_field.call_count, len(indexer.solr_fields))
        self.assertEqual(mock_schema.create_field.call_count, 0)

    @patch('arkindex.documents.indexer_v2.solr')
    def test_drop_index(self, mock_solr):
        indexer = Indexer('corpus_id')
        indexer.drop_index()

        self.assertEqual(mock_solr.delete_doc_by_query.call_count, 1)
        (index_name, query), kwargs = mock_solr.delete_doc_by_query.call_args
        self.assertEqual(index_name, indexer.collection_name)
        self.assertEqual(query, '*:*')
        self.assertDictEqual(kwargs, {'commit': True})

    def test_hash_worker(self):
        indexer = Indexer(None)
        self.assertIsNone(indexer.hash_worker(None))
        self.assertEqual(indexer.hash_worker(self.worker_version), f'{self.worker.id.hex[:10]}-{self.worker.name}')

    def test_build_element(self):
        annotated_pages = self.private_corpus.elements.filter(id=self.page.id).annotate(
            parent_id=Value(self.page.id, output_field=CharField()),
            parent_name=Value(self.page.name, output_field=CharField()),
            parent_type=Value(self.page.type.display_name, output_field=CharField()),
            type_name=Value(self.page.type.display_name, output_field=CharField())
        )
        indexer = Indexer(None)
        self.assertDictEqual(indexer.build_element(annotated_pages.first()), {
            'parent_id': str(self.page.id),
            'parent_name': self.page.name,
            'parent_type': self.page.type.display_name,
            'element_id': str(self.page.id),
            'element_text': self.page.name,
            'element_type': self.page.type.display_name,
            'element_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            'element_image': None
        })

    def test_build_transcriptions(self):
        tr_1 = self.page.transcriptions.create(
            confidence=0.8,
            text='Transcription on the page',
            worker_version=self.worker_version,
        )
        tr_2 = self.page.transcriptions.create(
            confidence=0.5,
            text='Second transcription',
            worker_version=self.worker_version,
        )
        indexer = Indexer(None)
        self.assertListEqual(indexer.build_transcriptions(self.page, {'key': 'value'}), [
            {
                'key': 'value',
                'transcription_id': str(tr_1.id),
                'transcription_confidence': tr_1.confidence,
                'transcription_text': tr_1.text,
                'transcription_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}'
            },
            {
                'key': 'value',
                'transcription_id': str(tr_2.id),
                'transcription_confidence': tr_2.confidence,
                'transcription_text': tr_2.text,
                'transcription_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}'
            }
        ])

    def test_build_classifications(self):
        cl_1 = self.page.classifications.create(
            confidence=0.8,
            worker_version=self.worker_version,
            ml_class=self.private_corpus.ml_classes.create(name='Cat')
        )
        cl_2 = self.page.classifications.create(
            confidence=0.4,
            worker_version=self.worker_version,
            ml_class=self.private_corpus.ml_classes.create(name='Dog')
        )
        indexer = Indexer(None)
        self.assertListEqual(indexer.build_classifications(self.page, {'key': 'value'}), [
            {
                'key': 'value',
                'classification_id': str(cl_1.id),
                'classification_name': cl_1.ml_class.name,
                'classification_confidence': cl_1.confidence,
                'classification_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            },
            {
                'key': 'value',
                'classification_id': str(cl_2.id),
                'classification_name': cl_2.ml_class.name,
                'classification_confidence': cl_2.confidence,
                'classification_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            }
        ])

    def test_build_metadatas(self):
        self.private_corpus.allowed_metadatas.create(type=MetaType.Location, name='Country')
        self.private_corpus.allowed_metadatas.create(type=MetaType.Text, name='Folio')
        md_1 = self.page.metadatas.create(
            type=MetaType.Location,
            name='Country',
            value='France',
            worker_version=self.worker_version,
        )
        md_2 = self.page.metadatas.create(
            type=MetaType.Text,
            name='Folio',
            value='1',
            worker_version=self.worker_version,
        )
        indexer = Indexer(None)
        self.assertListEqual(indexer.build_metadatas(self.page, {'key': 'value'}), [
            {
                'key': 'value',
                'metadata_id': str(md_1.id),
                'metadata_name': md_1.name,
                'metadata_text': md_1.value,
                'metadata_type': md_1.type.value,
                'metadata_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            },
            {
                'key': 'value',
                'metadata_id': str(md_2.id),
                'metadata_name': md_2.name,
                'metadata_text': md_2.value,
                'metadata_type': md_2.type.value,
                'metadata_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            }
        ])

    def test_build_entities(self):
        self.private_corpus.allowed_metadatas.create(type=MetaType.Location, name='Country')
        entity_1 = self.private_corpus.entities.create(name="CDLK", type=EntityType.Location, worker_version=self.worker_version)
        tr = self.page.transcriptions.create(
            confidence=0.8,
            text='Transcription on the page',
        )
        entity_1.transcription_entities.create(
            transcription=tr,
            offset=0,
            length=len(entity_1.name)
        )
        entity_2 = self.private_corpus.entities.create(name="Robert", type=EntityType.Person, worker_version=self.worker_version)
        self.page.metadatas.create(
            type=MetaType.Location,
            name='Country',
            value='France',
            entity=entity_2
        )
        indexer = Indexer(None)
        self.assertListEqual(indexer.build_entities(self.page, {'key': 'value'}), [
            {
                'key': 'value',
                'entity_id': str(entity_1.id),
                'entity_text': entity_1.name,
                'entity_type': entity_1.type.value,
                'entity_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            },
            {
                'key': 'value',
                'entity_id': str(entity_2.id),
                'entity_text': entity_2.name,
                'entity_type': entity_2.type.value,
                'entity_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            }
        ])

    @patch('arkindex.documents.indexer_v2.solr')
    def test_index(self, mock_solr):
        zone = Zone.objects.first()
        self.page.zone = zone
        self.page.save()

        entity = self.private_corpus.entities.create(name="CDLK", type=EntityType.Location, worker_version=self.worker_version)
        tr = self.page.transcriptions.create(
            confidence=0.8,
            text='Transcription on the page',
        )
        entity.transcription_entities.create(
            transcription=tr,
            offset=0,
            length=len(entity.name)
        )

        indexer = Indexer(self.private_corpus.id)
        with self.assertExactQueries('indexer_prefetch_v2.sql', params={
            'corpus_id': self.private_corpus.id,
            'page_id': self.page.id,
            'zone_id': zone.id,
            'image_id': zone.image.id,
            'worker_version_id': self.worker_version.id,
            'worker_id': self.worker.id,
            'transcription_id': tr.id
        }):
            indexer.index()
        self.assertEqual(mock_solr.index.call_count, 1)
        (index_name, documents), kwargs = mock_solr.index.call_args
        self.assertEqual(index_name, indexer.collection_name)
        self.assertListEqual(documents, [
            {
                'element_id': str(self.page.id),
                'element_text': self.page.name,
                'element_type': self.page.type.display_name,
                'element_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
                'element_image': zone.thumbnail_url,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.page.id),
                'element_text': self.page.name,
                'element_type': self.page.type.display_name,
                'element_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
                'element_image': zone.thumbnail_url,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'transcription_id': str(tr.id),
                'transcription_confidence': tr.confidence,
                'transcription_text': tr.text,
                'transcription_worker': None
            },
            {
                'element_id': str(self.page.id),
                'element_text': self.page.name,
                'element_type': self.page.type.display_name,
                'element_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
                'element_image': zone.thumbnail_url,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'entity_id': str(entity.id),
                'entity_text': entity.name,
                'entity_type': entity.type.value,
                'entity_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            }]
        )
        self.assertDictEqual(kwargs, {'commit': True})
