import uuid

from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus, Element
from arkindex.project.tests import FixtureAPITestCase


class TestParentsElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.private_elt = cls.private_corpus.elements.create(type=cls.private_corpus.types.create(slug='type'))
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def setUp(self):
        self.page = self.corpus.elements.get(name='Volume 1, page 1r')

    def test_element_parents(self):
        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1', ]
        )

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?folder')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1', ]
        )

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?folder=false')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?type=act')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        surface = Element.objects.get(name='Surface A')
        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(surface.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 1)
        self.assertEqual(response.json()['results'][0]['name'], 'Act 1')

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(surface.id)}) + '?recursive')
        self.assertEqual(len(response.json()['results']), 2)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Act 1', 'Volume 1'],
        )

        response = self.client.get(
            reverse('api:elements-parents', kwargs={'pk': str(surface.id)})
            + '?recursive&name=aCT'
        )
        self.assertEqual(len(response.json()['results']), 1)
        self.assertEqual(response.json()['results'][0]['name'], 'Act 1')

    def test_parents_with_corpus_false(self):
        response = self.client.get(
            reverse('api:elements-parents', kwargs={'pk': str(self.page.id)})
            + '?with_corpus=false'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(self.vol.id),
                "name": self.vol.name,
                "type": self.vol.type.slug,
                "worker_version_id": None,
                "zone": None,
                "best_classes": None,
                "has_children": None,
                "thumbnail_url": self.vol.thumbnail.s3_url,
            }
        ])

    def test_parents_worker_version(self):
        """
        List parents filtered by worker_version
        """
        self.corpus.elements.filter(name__contains='Volume 1').update(worker_version=self.worker_version)

        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'worker_version': str(self.worker_version.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1'],
        )

    def test_parents_worker_version_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'worker_version': 'blah'},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                data={'worker_version': uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})

    def test_parents_with_has_children(self):
        surface = Element.objects.get(name='Surface A')
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(surface.id)}),
                data={'recursive': True, 'with_has_children': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['has_children'] for element in response.json()['results']},
            {
                'Act 1': True,
                'Volume 1': True,
            }
        )

    def test_parents_modified_since_bad_format(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                HTTP_IF_MODIFIED_SINCE='AAAAAAAAAAAAAAAAA',
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'If-Modified-Since': ['Bad date format']})

    def test_parents_modified_since_not_modified(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_304_NOT_MODIFIED)

    def test_parents_modified_since(self):
        self.corpus.elements.filter(name='Volume 1').update(updated='2099-04-02T13:37:43Z')
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1', ]
        )
