from django.conf import settings
from django.db import OperationalError, models
from django.utils.functional import cached_property
from psycopg2.errors import ProgramLimitExceeded
from rest_framework.exceptions import APIException

# Lower-case markers that *may* indicate a IIIF server handles
# images after that path
IIIF_MARKERS = ('/iiif/', '/i3f/')


class ImageServerManager(models.Manager):

    @cached_property
    def local(self):
        return self.get(id=settings.LOCAL_IMAGESERVER_ID)

    def from_url(self, url):
        """
        Find a matching ImageServer from an image URL. Requires the server to be validated.
        May raise ImageServer.DoesNotExist or ImageServer.MultipleObjectsReturned.
        """
        servers = self \
            .annotate(image_url=models.Value(url, output_field=models.CharField())) \
            .filter(validated=True, image_url__startswith=models.F('url'))

        servers_count = servers.count()
        if not servers_count:

            # Try to detect a IIIF marker in the url
            # and create a new image server automatically
            for marker in IIIF_MARKERS:
                if marker not in url.lower():
                    continue
                server_url = url[:url.lower().index(marker) + len(marker)]
                return self.create(
                    url=server_url,
                    display_name=server_url,
                    validated=True,
                )

            raise self.model.DoesNotExist(
                f'The URL "{url}" does not match any existing image server. '
                'Please ask an instance administrator to register the IIIF server for this image.'
            )
        elif servers_count > 1:
            raise self.model.MultipleObjectsReturned(
                f'The URL "{url}" matched multiple existing image servers'
            )

        return servers.get()


class PolygonTooLargeError(APIException):
    status_code = 500
    default_detail = 'This polygon has exceeded the database size limits. Consider simplifying it.'
    default_code = 'polygon_too_large'


def _handle_polygon_error(method, *args, **kwargs):
    try:
        return method(*args, **kwargs)
    except OperationalError as e:
        if isinstance(e.__cause__, ProgramLimitExceeded) and len(e.args) > 0 and 'zone_unique_image_polygon' in e.args[0]:
            raise PolygonTooLargeError from e
        raise


class ZoneQuerySet(models.QuerySet):
    """
    This custom QuerySet handles index size overflow errors from the B-Tree index used by the
    zone_unique_image_polygon constraint.

    All the ways in which a model can be created or updated (.objects.create, .save, .objects.bulk_update,
    queryset.update(), …) all boil down to three methods: _insert, _update and update;
    this adds error handling to all of them.
    """

    def _insert(self, *args, **kwargs):
        return _handle_polygon_error(super()._insert, *args, **kwargs)

    def _update(self, *args, **kwargs):
        return _handle_polygon_error(super()._update, *args, **kwargs)

    def update(self, *args, **kwargs):
        return _handle_polygon_error(super().update, *args, **kwargs)


class ZoneManager(models.Manager):

    def get_queryset(self):
        return ZoneQuerySet(self.model, using=self._db)
