from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.project.tests import FixtureTestCase


class TestImageElements(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.img1 = cls.imgsrv.images.get(path='img1')

    def test_image_elements(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:image-elements', kwargs={'pk': self.img1.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertListEqual(
            [element['name'] for element in data['results']],
            ['Volume 1, page 1r', 'Surface A', 'Surface B', 'Text line', 'DATUM', 'PARIS', 'ROY'],
        )

    def test_image_elements_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:image-elements', kwargs={'pk': self.img1.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_image_elements_acl(self):
        """
        A user cannot list elements on corpus they have no guest access
        """
        self.client.force_login(self.user)
        private_corpus = Corpus.objects.create(name="Private", public=False)
        private_type = private_corpus.types.create(slug='cake', display_name='Cake')
        private_elt = private_corpus.elements.create(
            name='Moon',
            type=private_type,
            zone=self.img1.zones.create(polygon=[(100, 100), (142, 142), (133, 337), (100, 100)])
        )
        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:image-elements', kwargs={'pk': self.img1.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 7)
        response_names = [e['name'] for e in results]
        self.assertTrue('Volume 1, page 1r' in response_names)
        self.assertFalse(private_elt.name in response_names)

    def test_image_elements_type_filter(self):
        self.client.force_login(self.user)
        # Create an element of a different type with a zone
        cake_type = self.corpus.types.create(slug='cake', display_name='Cake')
        self.corpus.elements.create(
            type=cake_type,
            name='Quatre-quarts',
            zone=self.img1.zones.create(polygon=[(0, 0), (42, 42), (13, 37), (0, 0)]),
        )

        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:image-elements', kwargs={'pk': self.img1.id}) + '?type=cake')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertListEqual(
            [element['name'] for element in data['results']],
            ['Quatre-quarts'],
        )

    def test_image_elements_folder_filter(self):
        self.client.force_login(self.user)
        # Add a zone on a folder
        vol = self.corpus.elements.get(name='Volume 1')
        self.assertTrue(vol.type.folder)
        vol.zone = self.img1.zones.create(polygon=[(0, 0), (0, 1), (1, 1), (0, 0)])
        vol.save()

        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:image-elements', kwargs={'pk': self.img1.id}) + '?folder')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertListEqual(
            [element['name'] for element in data['results']],
            ['Volume 1'],
        )

    def test_image_elements_order_id_fallback(self):
        """
        Elements with the same type and name should be ordered by ID
        """
        self.client.force_login(self.user)
        elt_type = self.corpus.types.create(slug='duplicated', display_name='Duplicated')
        elts = Element.objects.bulk_create([
            Element(
                corpus=self.corpus,
                type=elt_type,
                name='Duplicata',
                zone=self.img1.zones.create(polygon=[(i, i), (i, 200), (200, 200), (200, i), (i, i)])
            ) for i in range(40)
        ])
        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:image-elements', kwargs={'pk': self.img1.id}), {'type': 'duplicated'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 40)
        self.assertListEqual(
            [str(e.id) for e in Element.objects.filter(id__in=[e.id for e in elts]).order_by('id')[:20]],
            [e['id'] for e in data['results']]
        )
