from unittest import TestCase

from django.contrib.gis.geos import LinearRing, Point
from rest_framework.serializers import ValidationError

from arkindex.images.models import POLYGON_MAX_POINTS
from arkindex.project.serializer_fields import LinearRingField, PointField


class TestLinearRingSerializerField(TestCase):

    def test_pointfield_to_json(self):
        self.assertListEqual(
            PointField().to_representation(Point(13, 37)),
            [13, 37],
        )

    def test_json_to_pointfield(self):
        self.assertEqual(
            PointField().run_validation([13, 37]),
            Point(13, 37),
        )

    def test_bad_json_to_pointfield(self):
        with self.assertRaisesRegex(ValidationError, 'Ensure this field has at least 2 elements.'):
            PointField().run_validation([])
        with self.assertRaisesRegex(ValidationError, 'Ensure this field has no more than 2 elements.'):
            PointField().run_validation([1, 2, 3])

        with self.assertRaises(ValidationError) as context:
            PointField().run_validation(['a', 5])
        self.assertDictEqual(
            {k: list(map(str, v)) for k, v in context.exception.detail.items()},
            {0: ['A valid integer is required.']}
        )

    def test_linear_ring_to_json(self):
        self.assertListEqual(
            LinearRingField().to_representation(LinearRing(
                Point(-40, 50),
                Point(10, 20),
                Point(30, 40),
                Point(-40, 50),
            )),
            [
                [-40, 50],
                [10, 20],
                [30, 40],
                [-40, 50],
            ]
        )

    def test_json_to_linear_ring(self):
        self.assertTupleEqual(
            LinearRingField().run_validation([
                [30, 40],
                [10, 20],
                [10, 20],
                [-40, 50],
                [30, 40],
            ]).coords, LinearRing(
                Point(-40, 50),
                Point(30, 40),
                Point(10, 20),
                Point(-40, 50),
            ).coords
        )

    def test_bad_json_to_linear_ring(self):
        with self.assertRaisesRegex(ValidationError, 'Ensure this field has at least 3 elements.'):
            LinearRingField().run_validation([])

        with self.assertRaises(ValidationError) as context:
            LinearRingField().run_validation([1])
        self.assertDictEqual(
            {k: list(map(str, v)) for k, v in context.exception.detail.items()},
            {0: ['Expected a list of items but got type "int".']}
        )

        with self.assertRaisesRegex(ValidationError, 'Ensure this field has at least 3 elements.'):
            LinearRingField().run_validation([[1, 2]])

    def test_linear_ring_max_size(self):
        """
        Polygon maximum allowed number of points is empirically set to POLYGON_MAX_POINTS
        """
        # Create a polygon with POLYGON_MAX_POINTS - 1 points
        coords = [(0, 0), *((i, i) for i in range(POLYGON_MAX_POINTS - 2, 0, -1))]
        self.assertEqual(len(coords), POLYGON_MAX_POINTS - 1)
        self.assertEqual(
            LinearRingField().run_validation(coords).coords,
            LinearRing(coords + [[0, 0]]).coords,
        )

        # Next point makes the validation fail if it is not equal to the first one
        coords.append([POLYGON_MAX_POINTS + 1, POLYGON_MAX_POINTS + 1])
        with self.assertRaisesRegex(ValidationError, f'{POLYGON_MAX_POINTS} points are allowed only when the first and last points are equal.'):
            LinearRingField().run_validation(coords)

        coords[-1] = [0, 0]
        self.assertEqual(
            LinearRingField().run_validation(coords).coords,
            LinearRing(coords).coords,
        )
